﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.ALPS.SteadyState {
  [Item("AlpsSsMover", "")]
  [StorableClass]
  public class AlpsSsMover : SingleSuccessorOperator, IStochasticOperator {
    private ILookupParameter<IntValue> TargetIndexParameter {
      get { return (ILookupParameter<IntValue>)Parameters["TargetIndex"]; }
    }
    private ILookupParameter<IntValue> PopulationSizeParameter {
      get { return (ILookupParameter<IntValue>)Parameters["PopulationSize"]; }
    }
    private ILookupParameter<IntValue> NumberOfLayersParameter {
      get { return (ILookupParameter<IntValue>)Parameters["NumberOfLayers"]; }
    }
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }

    [StorableConstructor]
    private AlpsSsMover(bool deserializing)
      : base(deserializing) { }
    private AlpsSsMover(AlpsSsMover original, Cloner cloner)
      : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new AlpsSsMover(this, cloner);
    }

    public AlpsSsMover()
      : base() {
      Parameters.Add(new LookupParameter<IntValue>("TargetIndex"));
      Parameters.Add(new LookupParameter<IntValue>("PopulationSize"));
      Parameters.Add(new LookupParameter<IntValue>("NumberOfLayers"));
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));
    }

    public override IOperation Apply() {
      int i = ((IntValue)ExecutionContext.Scope.Variables["Layer"].Value).Value;
      int j = TargetIndexParameter.ActualValue.Value;
      int n = NumberOfLayersParameter.ActualValue.Value;
      int m = PopulationSizeParameter.ActualValue.Value;
      var random = RandomParameter.ActualValue;

      var layer = ExecutionContext.Scope.SubScopes[0];
      var newIndividual = ExecutionContext.Scope.SubScopes[1];
      var gs = ExecutionContext.Scope.Parent;

      if (i < n) {
        TryMoveUp(gs, newIndividual, layer, i, j, n, m, random);
        layer.SubScopes[j] = newIndividual;
        newIndividual.Name = j.ToString();
      }

      return base.Apply();
    }

    private void TryMoveUp(IScope gs, IScope individual, IScope layer, int i, int j, int n, int m, IRandom random) {
      if (i < n) {
        if (gs.SubScopes.Count <= i + 1) {
          var newLayer = new Scope((i + 1).ToString());
          newLayer.Variables.Add(new Variable("Layer", new IntValue(i + 1)));
          newLayer.Variables.Add(new Variable("LayerPopulationSize", new IntValue(1)));
          gs.SubScopes.Add(newLayer);
        }

        var higherLayer = gs.SubScopes[i + 1];
        var replaceIndex = FindReplaceable(higherLayer, individual, random, m);
        if (replaceIndex.HasValue) {
          var replacedIndividual = layer.SubScopes[j];
          if (replaceIndex.Value >= higherLayer.SubScopes.Count) {
            higherLayer.SubScopes.Add(replacedIndividual);
          } else {
            TryMoveUp(gs, higherLayer.SubScopes[j], higherLayer, i + 1, replaceIndex.Value, n, m, random);
            // TODO higherlayer[j] does not exist
            higherLayer.SubScopes[replaceIndex.Value] = replacedIndividual;
          }
        }
      }
    }

    private int? FindReplaceable(IScope layer, IScope individual, IRandom random, int m) {
      if (layer.SubScopes.Count < m)
        return layer.SubScopes.Count;
      return Enumerable.Range(0, layer.SubScopes.Count).SampleRandom(random);
    }
  }
}